import os
import argparse
import numpy as np
import warnings
import pickle
import glob
import obspy
import random

from tqdm import tqdm
from torch.utils.data import Dataset

warnings.filterwarnings('ignore')

class ModelNetDataLoader(Dataset):
    def __init__(self, root, args, split='train', process_data=False):
        self.root = root
        self.npoints = args.num_point
        self.process_data = process_data
        self.num_category = args.num_category

        # self.cat = ['earthquakes','glacial']
        # self.classes = dict(zip(self.cat, range(len(self.cat))))

        EarthBinaryDataDirectory = os.path.join(self.root,'earthquakes')
        EarthBinaryFileNames = glob.glob(EarthBinaryDataDirectory+'/*.bin')

        random.shuffle(EarthBinaryFileNames)
        EarthTrain = EarthBinaryFileNames[0:500]
        EarthTest = EarthBinaryFileNames[500:]

        # shape_ids = {}
        # shape_ids['train'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet10_train.txt'))]
        # shape_ids['test'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet10_test.txt'))]

        assert (split == 'train' or split == 'test')
        self.datapath = []
        if split=='train':
            for i in range(len(EarthTrain)):
                MseedFileName = EarthTrain[i][:-4] + '.mseed'
                self.datapath = self.datapath + [(MseedFileName, EarthTrain[i])]
        elif split=='test':
            for i in range(len(EarthTest)):
                MseedFileName = EarthTest[i][:-4] + '.mseed'
                self.datapath = self.datapath + [(MseedFileName, EarthTest[i])]

        print('The size of %s data is %d' % (split, len(self.datapath)))

        self.save_path = os.path.join(root, 'modelnet_predict_earth%d_%s_%dpts.dat' % (self.num_category, split, self.npoints))

        if self.process_data:
            if not os.path.exists(self.save_path):
                print('Processing data %s (only running in the first time)...' % self.save_path)
                self.list_of_points = [None] * len(self.datapath)
                self.list_of_labels = [None] * len(self.datapath)

                for index in tqdm(range(len(self.datapath)), total=len(self.datapath)):
                    fn = self.datapath[index]
                    cls = np.zeros((4,))
                    DataStream = obspy.read(fn[0],format='MSEED')
                    StartString = str(DataStream[0].stats.starttime)[0:-11]
                    with open("./earthquakes_catalog.txt", 'r') as f:
                        for line in f:
                            if line.startswith(StartString):
                                SourceLat = float(line.split('|')[1].split(',')[0])
                                SourceLon = float(line.split('|')[1].split(',')[1])
                                paramList = line.split()
                                cls = np.array([SourceLat/180.0, SourceLon/180.0, float(paramList[5]), float(paramList[8])])
                                break
                    cls = cls.astype(np.float32)
                    # DataArrayRead = np.fromfile(BinaryFileName,dtype='float32').reshape(129,7200)
                    point_set = np.fromfile(fn[1],dtype='float64').reshape(129,7200)
                    # point_set = np.loadtxt(fn[1], delimiter=',').astype(np.float32)

                    # 去除NaN
                    point_set = np.where(np.isnan(point_set), 0, point_set)
                    # # 高斯噪声
                    # point_set = point_set + np.random.normal(0,0.012,129*7200).reshape(129,7200)
                    # 平滑
                    # for i in range(7200):
                    #     point_set[:,i] = np.max(point_set[:,np.max([0,i-5]):np.min([7200,i+6])])

                    StationLatLon = dict()
                    pos_set = np.zeros((129,2))
                    tempi = 0
                    with open("./stations_full.txt", 'r') as f:
                        for line in f:
                            if line.startswith('#'):
                                continue
                            NetworkCode = line.split()[0]
                            StationCode = line.split()[1]
                            StationLatLon[NetworkCode+'.'+StationCode] = [float(line.split()[2]),float(line.split()[3])]
                            pos_set[tempi] = np.array([float(line.split()[2]),float(line.split()[3])])
                            tempi += 1

                    # # 高斯噪声
                    # pos_set = pos_set + np.random.normal(0,0.01,129*2).reshape(129,2)

                    point_full = np.concatenate((pos_set,point_set),axis=1)
                    point_set = point_full
                    point_set = point_set[0:self.npoints, :]

                    self.list_of_points[index] = point_set
                    self.list_of_labels[index] = cls

                with open(self.save_path, 'wb') as f:
                    pickle.dump([self.list_of_points, self.list_of_labels], f)
            else:
                print('Load processed data from %s...' % self.save_path)
                with open(self.save_path, 'rb') as f:
                    self.list_of_points, self.list_of_labels = pickle.load(f)

    def __len__(self):
        return len(self.datapath)

    def _get_item(self, index):
        if self.process_data:
            point_set, label = self.list_of_points[index], self.list_of_labels[index]
        else:
            fn = self.datapath[index]
            cls = np.zeros((4,))
            DataStream = obspy.read(fn[0],format='MSEED')
            StartString = str(DataStream[0].stats.starttime)[0:-11]
            with open("./earthquakes_catalog.txt", 'r') as f:
                for line in f:
                    if line.startswith(StartString):
                        SourceLat = float(line.split('|')[1].split(',')[0])
                        SourceLon = float(line.split('|')[1].split(',')[1])
                        paramList = line.split()
                        cls = np.array([SourceLat/180.0, SourceLon/180.0, float(paramList[5]), float(paramList[8])])
                        break
            label = cls.astype(np.float32)
            point_set = np.fromfile(fn[1],dtype='float64').reshape(129,7200)

            point_set = np.where(np.isnan(point_set), 0, point_set)
            # # 高斯噪声
            # point_set = point_set + np.random.normal(0,0.012,129*7200).reshape(129,7200)

            StationLatLon = dict()
            pos_set = np.zeros((129,2))
            tempi = 0
            with open("./stations_full.txt", 'r') as f:
                for line in f:
                    if line.startswith('#'):
                        continue
                    NetworkCode = line.split()[0]
                    StationCode = line.split()[1]
                    StationLatLon[NetworkCode+'.'+StationCode] = [float(line.split()[2]),float(line.split()[3])]
                    pos_set[tempi] = np.array([float(line.split()[2]),float(line.split()[3])])
                    tempi += 1

            # # 高斯噪声
            # pos_set = pos_set + np.random.normal(0,0.01,129*2).reshape(129,2)

            point_full = np.concatenate((pos_set,point_set),axis=1)
            point_set = point_full
            point_set = point_set[0:self.npoints, :]
                
        point_set[:, 0:2] = (point_set[:, 0:2]) / 180.0

        return point_set, label
    def __getitem__(self, index):
        return self._get_item(index)


# for tuning and debugging usage
"""
def parse_args():
    '''PARAMETERS'''
    parser = argparse.ArgumentParser('training')
    parser.add_argument('--use_cpu', action='store_true', default=False, help='use cpu mode')
    parser.add_argument('--gpu', type=str, default='0', help='specify gpu device')
    parser.add_argument('--batch_size', type=int, default=4, help='batch size in training')
    parser.add_argument('--model', default='geonet_predict_simple_earth', help='model name [default: geonet_predict_simple_earth]')
    parser.add_argument('--num_category', default=2, type=int, choices=[2, 3],  help='how many different categories in dataset')
    parser.add_argument('--epoch', default=200, type=int, help='number of epoch in training')
    parser.add_argument('--learning_rate', default=0.001, type=float, help='learning rate in training')
    parser.add_argument('--num_point', type=int, default=129, help='Point Number')
    parser.add_argument('--optimizer', type=str, default='Adam', help='optimizer for training')
    parser.add_argument('--log_dir', type=str, default=None, help='experiment root')
    parser.add_argument('--decay_rate', type=float, default=1e-4, help='decay rate')
    parser.add_argument('--num_channel', type=int, default=7202, help='channels Number')
    parser.add_argument('--process_data', action='store_true', default=False, help='save data offline')
    return parser.parse_args()

if __name__ == '__main__':
    import torch

    args = parse_args()
    data = ModelNetDataLoader('data/',args, split='train')
    DataLoader = torch.utils.data.DataLoader(data, batch_size=20, shuffle=True)
    for point, label in DataLoader:
        print(point.shape)
        print(label.shape)
        print(label)
"""