import os
import sys
import torch
import numpy as np

import datetime
import logging
import importlib
import shutil
import argparse
import GeoDataloader2_earth

from pathlib import Path
from tqdm import tqdm

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = BASE_DIR

def parse_args():
    '''PARAMETERS'''
    parser = argparse.ArgumentParser('training')
    parser.add_argument('--use_cpu', action='store_true', default=False, help='use cpu mode')
    parser.add_argument('--gpu', type=str, default='0', help='specify gpu device')
    parser.add_argument('--batch_size', type=int, default=4, help='batch size in training')
    parser.add_argument('--model', default='geonet_predict_simple_earth', help='model name [default: geonet_predict_simple_earth]')
    parser.add_argument('--num_category', default=4, type=int,  help='how many different categories in dataset')
    parser.add_argument('--epoch', default=200, type=int, help='number of epoch in training')
    parser.add_argument('--learning_rate', default=7e-4, type=float, help='learning rate in training')
    parser.add_argument('--num_point', type=int, default=129, help='Point Number')
    parser.add_argument('--optimizer', type=str, default='Adam', help='optimizer for training')
    parser.add_argument('--log_dir', type=str, default=None, help='experiment root')
    parser.add_argument('--decay_rate', type=float, default=1e-3, help='decay rate')
    parser.add_argument('--num_channel', type=int, default=7202, help='channels Number')
    parser.add_argument('--process_data', action='store_true', default=False, help='save data offline')
    return parser.parse_args()


def inplace_relu(m):
    classname = m.__class__.__name__
    if classname.find('ReLU') != -1:
        m.inplace=True

def random_point_dropout(batch_pc, max_dropout_ratio=0.25):
    ''' 
    batch_pc: BxNx3 
    对batch中每一个数据选取一部分点来去掉,用第一个点来替代
    '''
    for b in range(batch_pc.shape[0]):
        dropout_ratio =  np.random.random()*max_dropout_ratio # 0~0.875
        drop_idx = np.where(np.random.random((batch_pc.shape[1]))<=dropout_ratio)[0]
        if len(drop_idx)>0:
            batch_pc[b,drop_idx,:] = batch_pc[b,0,:] # set to the first point
    return batch_pc


def determine_correct_rate_earth(pred, target):
    B, _ = pred.size()
    delta = pred - target
    sum = 0
    for i in range(B):
        if(abs(delta[i,0]) * 180.0 > 2): continue
        # if(abs(delta[i,1]) * 180.0 > 7): continue
        # if(abs(delta[i,2]) > 0.1): continue
        # if(abs(delta[i,3]) > 100): continue
        sum += 1
    return float(sum/B)

def determine_correct_more_earth(pred, target):
    B, _ = pred.size()
    delta = pred - target
    sum = 0
    for i in range(B):
        if(abs(delta[i,0]) * 180.0 > 0.7): continue
        # if(abs(delta[i,1]) * 180.0 > 0.7): continue
        # if(abs(delta[i,2]) > 0.1): continue
        # if(abs(delta[i,3]) > 100): continue
        sum += 1
    return float(sum/B)

def determine_correct_bad_earth(pred, target):
    B, _ = pred.size()
    delta = pred - target
    sum = 0
    for i in range(B):
        if(abs(delta[i,0]) * 180.0 > 5): continue
        # if(abs(delta[i,1]) * 180.0 > 12): continue
        # if(abs(delta[i,2]) > 0.1): continue
        # if(abs(delta[i,3]) > 100): continue
        sum += 1
    return float(sum/B)


def test(model, loader, num_class=40):
    mean_correct = []
    more_correct = []
    bad_correct = []
    classifier = model.eval()

    for j, (points, target) in tqdm(enumerate(loader), total=len(loader)):
        if not args.use_cpu:
            points, target = points.cuda(), target.cuda()
        points = points.transpose(2, 1)
        points = points.type(torch.FloatTensor)
        if not args.use_cpu:
            points = points.cuda()
        pred = classifier(points)

        if(j==0):
            print(pred[0])
            print(target[0])

        mean_correct.append(determine_correct_rate_earth(pred,target))
        more_correct.append(determine_correct_more_earth(pred,target))
        bad_correct.append(determine_correct_bad_earth(pred,target))

    instance_acc = np.mean(mean_correct)
    more_acc = np.mean(more_correct)
    bad_acc = np.mean(bad_correct)
    return instance_acc, more_acc, bad_acc


def main(args):
    def log_string(str):
        logger.info(str)
        print(str)

    '''HYPER PARAMETER'''
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu

    '''CREATE DIR'''
    timestr = str(datetime.datetime.now().strftime('%Y-%m-%d_%H-%M'))
    exp_dir = Path('./log/')
    exp_dir.mkdir(exist_ok=True)
    exp_dir = exp_dir.joinpath('classification')
    exp_dir.mkdir(exist_ok=True)
    if args.log_dir is None:
        exp_dir = exp_dir.joinpath(timestr)
    else:
        exp_dir = exp_dir.joinpath(args.log_dir)
    exp_dir.mkdir(exist_ok=True)
    checkpoints_dir = exp_dir.joinpath('checkpoints/')
    checkpoints_dir.mkdir(exist_ok=True)
    log_dir = exp_dir.joinpath('logs/')
    log_dir.mkdir(exist_ok=True)

    '''LOG'''
    args = parse_args()
    logger = logging.getLogger("Model")
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler('%s/%s.txt' % (log_dir, args.model))
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    log_string('PARAMETER ...')
    log_string(args)

    '''DATA LOADING'''
    log_string('Load dataset ...')
    data_path = 'data/'

    train_dataset = GeoDataloader2_earth.ModelNetDataLoader(root=data_path, args=args, split='train', process_data=args.process_data)
    test_dataset = GeoDataloader2_earth.ModelNetDataLoader(root=data_path, args=args, split='test', process_data=args.process_data)
    trainDataLoader = torch.utils.data.DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, num_workers=0, drop_last=True)
    testDataLoader = torch.utils.data.DataLoader(test_dataset, batch_size=args.batch_size, shuffle=False, num_workers=0)

    '''MODEL LOADING'''
    num_class = args.num_category
    model = importlib.import_module(args.model)
    shutil.copy('./%s.py' % args.model, str(exp_dir))
    shutil.copy('./geonet_util.py', str(exp_dir))
    shutil.copy('./train_cls.py', str(exp_dir))

    classifier = model.get_model(num_class, args.num_channel)
    criterion = model.get_loss()
    classifier.apply(inplace_relu)

    if not args.use_cpu:
        classifier = classifier.cuda()
        criterion = criterion.cuda()

    try:
        checkpoint = torch.load(str(exp_dir) + '/checkpoints/best_model.pth')
        start_epoch = checkpoint['epoch']
        classifier.load_state_dict(checkpoint['model_state_dict'])
        log_string('-------Use pretrain model-------')
        log_string('current performance:')
        acc1 = test(classifier.eval(), trainDataLoader, num_class=num_class)
        acc2 = test(classifier.eval(), testDataLoader, num_class=num_class)
        print(acc1,"----",acc2)
        log_string('-------Use pretrain model-------')

    except:
        log_string('-------No existing model, starting training from scratch...-------')
        start_epoch = 0

    if args.optimizer == 'Adam':
        optimizer = torch.optim.Adam(
            classifier.parameters(),
            lr=args.learning_rate,
            betas=(0.9, 0.999),
            eps=1e-08,
            weight_decay=args.decay_rate
        )
    else:
        optimizer = torch.optim.SGD(classifier.parameters(), lr=0.01, momentum=0.9)

    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.7)
    global_epoch = 0
    global_step = 0
    best_instance_acc = 0.0
    best_class_acc = 0.0

    '''TRANING'''
    logger.info('Start training...')
    for epoch in range(start_epoch, args.epoch):
        log_string('Epoch %d (%d/%s):' % (global_epoch + 1, epoch + 1, args.epoch))
        mean_correct = []
        mean_loss = []
        classifier = classifier.train()

        scheduler.step()
        for batch_id, (points, target) in tqdm(enumerate(trainDataLoader, 0), total=len(trainDataLoader), smoothing=0.9):
        # batch_id = 1
        # for points, target in trainDataLoader:
        #     print("\nbatch ",batch_id, " under dealing...\n")
        #     batch_id += 1
        #     print("batchid: ",batch_id,"\nsum: ", points.sum())
            optimizer.zero_grad()

            points = points.data.numpy()
            points = random_point_dropout(points)
            # points[:, :, 0:3] = provider.random_scale_point_cloud(points[:, :, 0:3])
            # points[:, :, 0:3] = provider.shift_point_cloud(points[:, :, 0:3])
            points = torch.Tensor(points)
            points = points.transpose(2, 1)
            # print(points.size())

            if not args.use_cpu:
                points, target = points.cuda(), target.cuda()

            pred = classifier(points)
            # print("-----",pred.size())
            # print("\n",pred,"\n")
            loss = criterion(pred, target)
            # print("\n",loss,"\n")

            batch_loss = loss.cpu()
            mean_loss.append(batch_loss.detach().numpy())
            mean_correct.append(determine_correct_rate_earth(pred, target))
            loss.backward()
            optimizer.step()
            global_step += 1

        train_loss = np.mean(mean_loss)
        train_correct_acc = np.mean(mean_correct)
        log_string('Train Loss: %f' % train_loss)
        log_string('Train Correct Acc: %f' % train_correct_acc)

        with torch.no_grad():
            instance_acc, more_acc, bad_acc = test(classifier.eval(), testDataLoader, num_class=num_class)

            if (instance_acc >= best_instance_acc):
                best_instance_acc = instance_acc
                best_epoch = epoch + 1

            log_string('Bad  Instance Accuracy: %f' % (bad_acc))
            log_string('Test Instance Accuracy: %f' % (instance_acc))
            log_string('More Instance Accuracy: %f' % (more_acc))
            log_string('Best Instance Accuracy: %f' % (best_instance_acc))

            if (instance_acc >= best_instance_acc):
                logger.info('Save model...')
                savepath = str(checkpoints_dir) + '/best_model.pth'
                log_string('Saving at %s' % savepath)
                state = {
                    'epoch': best_epoch,
                    'instance_acc': instance_acc,
                    'model_state_dict': classifier.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                }
                torch.save(state, savepath)
            global_epoch += 1

    logger.info('End of training...')


if __name__ == '__main__':
    args = parse_args()
    main(args)